' File Name: TextBox.bas
' Versioin:  1.0
' Date       11/6/2000
'
' Written by tshoemaker@yahoo.com
'
' This program lets the user select a text or text node elemant and
' places a box around it.  The box has a "padded" size of th/2.
'
' This program uses code from locate.bas from Bentley.  See
' info below for details.
'
' Revision History
' 11/06/200  TCS Program Written
' 08/03/2001  - Chuck Houston
'               Gulf Interstate Engineering
'             - Change Locate Command to exit without error with a
'               Command or Keyin.
'             - Create a graphic group from text and box
'
' This software is provided  "AS IS",  WITHOUT  WARRANTY OF ANY
' KIND.  THE AUTHOR AND THE COPYRIGHT HOLDER OF THIS SOFTWARE
' DISCLAIM ALL WARRANTIES OF ANY KIND, EITHER EXPRESS OR IMPLIED,
' INCLUDING, WITHOUT LIMITATION, IMPLIED WARRANTIES OF
' MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
'--------------------------------------------------------------------
' Example of element location technique
'--------------------------------------------------------------------
'--------------------------------------------------------------------
'
'  Copyright (1995) Bentley Systems, Inc., All rights reserved.
'
'    $RCSfile: locate.bas,v $
'   $Revision: 1.1 $
'       $Date: 1998/06/18 20:03:09 $
'
'  "MicroStation" is a registered trademark of Bentley Systems, Inc.
'
'  Limited permission is hereby granted to reproduce and modify this
'  copyrighted material provided that the resulting code is used only
'  in conjunction with Bentley Systems products under the terms of the
'  license agreement provided therein, and that this notice is retained
'  in its entirety in any such reproduction or modification.
'
'--------------------------------------------------------------------
'
' This example uses a number of the MicroStation Basic Extension
' element location techniques. The actual manipulation that it
' performs on selected elements is very simple in this example, but
' the same technique could be used to implement a more sophisticated
' manipulation. Similarly, the filtering that it does during the
' element selection logic is simple, but the technique could be
' readily extended.
'
' This example looks fairly complicated because it emulates all of
' the behavior that MicroStation exhibits when it locates elements:
'
'  1. When you reject an element that is hilited, it keeps looking
'     around the data point for more elements that have not been
'     processed.
'  2. When you select an element and accept it with a point that
'     is close to the selection point, it looks for more
'     elements around the same point, skipping elements already
'     processed.
'  3. When you select an element and accept it with a point that
'     is not close to the selection point, it uses the accept
'     point to begin looking for more elements to process.
'
' In addition, it provides a step that filters elements so that only
' elements that are acceptable to the program are presented to the
' user for acceptance or rejection.
'
'--------------------------------------------------------------------

'---------------------------------------------------------------
'
'   locate_pointsCloseEnough - returns TRUE if points roughly
'               within the locate tolerance, FALSE otherwise
'
'---------------------------------------------------------------
Function locate_pointsCloseEnough (point1 as MbePoint, point2 as MbePoint) as Integer
    If Abs (point1.x - point2.x) + Abs (point1.y - point2.y) _
                    < MbeState.LocateTolerance Then
        locate_pointsCloseEnough = TRUE
    End If
End Function


'---------------------------------------------------------------
'
'   locate_manipulateElement - performs the actual manipulation
'               on the selected element. This manipulation
'               sequences commands and uses the MbeRelocate
'               subroutine to make the command work on the
'               previously selected element
'
'---------------------------------------------------------------
Sub locate_manipulateElement (point as MbePoint, view as Integer)

    Dim elem as New MbeElement
    Dim filePos as Long

    ' turn off graphics while changing element
    MbeState.noElementDisplay = 1

    ' start change color and relocate element
    'MbeSendCommand "CHANGE COLOR"
    'MbeRelocate point, view
    'MbeSendDataPoint point, view

    ' start change weight and relocate element
    'MbeSendCommand "CHANGE WEIGHT"
    'MbeRelocate point, view
    'MbeSendDataPoint point, view

    ' relocate the manipulated element to reset the "locate context",
    ' so if we continue the locate, we'll continue with the next element
    ' before we do so, we set the command to something harmless, so
    ' the data point sent by MbeRelocate doesn't have the effect of
    ' changing the weight of an element that may have been selected
    ' by sending the accept data point to the CHANGE WEIGHT command.
    ' (NOTE: this section could have been shortened by substituting
    '        MbeRelocate for the above MbeSendDataPoint, but this
    '        looks a bit confusing).
    MbeSendCommand "LOCELE"
    MbeRelocate point, view

    MbeState.noElementDisplay = 0

    ' get the element to redraw it with its new symbology
    filePos = elem.fromLocate()

    textBox elem
    'elem.display MBE_NormalDraw

    ' set MicroStation to a neutral state.
    MbeSendCommand "NULL"

End Sub


'---------------------------------------------------------------
'
'   locate_isElementAcceptable - returns TRUE if element passes
'                       filter criteria and FALSE otherwise.
'
'---------------------------------------------------------------
Function locate_isElementAcceptable (elem as MbeElement) as Integer
    If elem.type = MBE_Text or elem.type = MBE_TextNode Then
        locate_isElementAcceptable = TRUE
    End If
End Function

'---------------------------------------------------------------
'
'   locate_elementUnacceptable - reports an error saying that
'                       no acceptable element was located.
'
'---------------------------------------------------------------
Sub locate_elementUnacceptable
    Beep
    MbeWriteStatus "Only text or text node allowed"
End Sub

'---------------------------------------------------------------
'
'   locate_displayPrompts - tell the user what to identify
'
'---------------------------------------------------------------
Sub locate_displayPrompts
    MbeWritePrompt  "Identify Text"
    MbeWriteCommand "TextBox"
End Sub

'---------------------------------------------------------------
'
'   locate_selectElementsToModify - solicits user for elements to
'                       modify and manipulates them.
'
'---------------------------------------------------------------
Sub locate_selectElementsToModify
    Dim elem            as New MbeElement
    Dim filePos         as Long
    Dim point           as MbePoint
    Dim acceptPoint     as MbePoint
    Dim acceptScreenPt  as MbePoint
    Dim firstScreenPt   as MbePoint
    Dim stat            as Integer
    Dim view            as Integer
    Dim acceptView      as Integer
    Dim continueLocate  as Integer
    Dim haveAcceptPoint as Integer

    continueLocate      = FALSE
    haveAcceptPoint     = FALSE

    Do
        ' Start locating an element, don't put out msgs,
        ' don't hilite when found, don't allow components,
        ' continue locate from previous element if appropriate.
        MbeStartLocate 0,0,0,continueLocate
        MbeWritePrompt  "Identify Text"
        MbeWriteCommand "TextBox"

        ' If accepted element or continuing from a reset
        ' send the previously entered point.
        If haveAcceptPoint Or continueLocate Then
            Call MbeSendDataPoint (acceptPoint, acceptView)
            continueLocate = FALSE

            ' If previous element accepted, future comparisons are to accept point
            If haveAcceptPoint Then
                haveAcceptPoint = FALSE
                point = acceptPoint
                view  = acceptView
                firstScreenPt = acceptScreenPt
            End If
        Else
            ' Wait for a data point or a reset
            MbeGetInput MBE_DataPointInput, _
                        MBE_ResetInput,  _
                        MBE_CommandInput, _
                        MBE_KeyInInput

            Select Case MbeState.InputType
            Case MBE_DataPointInput
                stat = MbeState.getInputDataPoint (point, view, firstScreenPt)
                MbeSendLastInput
            Case MBE_ResetInput
                End
            Case MBE_CommandInput, MBE_KeyInInput
                MbeSendLastInput
                End
            End Select
        End If

        ' If we found an element, check for acceptability
        If MbeState.cmdResult = MBE_AcceptQuery Then
            ' if we did not find a line, keep resetting until
            '    either we find one or we run out of candidates
            Do
                filePos = elem.fromLocate ()
                If locate_isElementAcceptable (elem) Then

                    ' located an acceptable element - hilite it
                    elem.display MBE_Hilite

                    ' prompt the user to accept or reject it
                    MbeWritePrompt "Accept / Reject"

                    ' get data to accept, reset to reject
                    MbeGetInput MBE_DataPointInput, MBE_ResetInput

                    If MbeState.inputType = MBE_DataPointInput Then
                        ' on Data point, accept element and modify it
                        stat = MbeState.getInputDataPoint (acceptPoint, _
                                                    acceptView, acceptScreenPt)
                        Call locate_manipulateElement (point, view)

                        ' continue locating if accept point near first point
                        continueLocate = locate_pointsCloseEnough (acceptScreenPt, _
                                                                   firstScreenPt)
                        haveAcceptPoint = TRUE
                    Else
                        ' user rejected it - continue looking around the same point
                        ' by continuing the locate operation in the loop.
                        elem.display MBE_NormalDraw
                        acceptPoint = point
                        acceptView = view
                        continueLocate = TRUE
                    End If
                    Exit Do
                Else
                    ' found an element but it is not acceptable to our filter.
                    ' Send resets to the locate logic to cause it to retrieve
                    ' the next closest element in range
                    MbeSendReset
                    If MbeState.cmdResult = MBE_ElementNotFound Then
                        Call locate_elementUnacceptable
                        Exit Do
                    End If
                End If
            Loop While elem.type <> MBE_Text and elem.type <> MBE_TextNode
        ElseIf MbeState.cmdResult = MBE_ElementNotFound Then
            MbeWriteError "Element Not Found"
        Else
            print "unexpected MbeState.cmdResult of "; MbeState.cmdResult
        End If
    Loop

End Sub

'---------------------------------------------
'This is where the box is calculated and drawn
Sub textBox( elem as mbeElement )
  dim endOfFilePos  as Long
  dim filePos       as Long
  dim range as MbeRange
  dim status as long

  'variables for rotation
  Dim rotation(1 to 3, 1 to 3) as Double
  Dim tmp( 1 to 3, 1 to 3 ) as Double
  Dim r1 as Double
  Dim r2 as Double
  Dim txang as Double

  'variables for conv. from uor to mu
  Dim conv as double
  Dim globalOrg as MbePoint

  'variables for range box
  Dim x1 as double
  Dim x2 as double
  Dim y1 as double
  Dim y2 as double

  'variables for figuring text box
  Dim boxH as Double
  Dim nLines as Integer
  Dim org as MbePoint
  Dim Txt as String
  Dim txH as Double
  Dim p1 as MbePoint
  Dim p2 as MbePoint
  Dim p3 as MbePoint
  Dim p4 as MbePoint

  'temp variables
  Dim t1 as Double
  Dim t2 as Double
  Dim t3 as Double
  Dim t4 as Double

  conv = MbeDgnInfo.uorPerSub * MbeDgnInfo.subPerMaster
  status = MbeDgnInfo.getGlobalOrigin(globalOrg)

  globalOrg.X = -globalOrg.X/conv
  globalOrg.Y = -globalOrg.Y/conv

  'if elem is part of textnode set to header
  status = elem.headerElement

  txH = elem.CharHeight
  If elem.getRange(range) = MBE_Success Then
    x1 = range.xLow  / conv + globalOrg.X
    x2 = range.xHigh / conv + globalOrg.X
    y1 = range.yLow  / conv + globalOrg.Y
    y2 = range.yHigh / conv + globalOrg.Y
    If elem.getRotation(rotation) = MBE_Success Then
      r1 = rotation(1,1) 'acos
      r2 = rotation(2,1) 'asin
      If (r1=-1)  or (r1=1) Then
        txang = 0
        If (r1=-1) Then txang = PI
      Else
        txang = (Atn(-r1/Sqr(-r1*r1+1))+2*Atn(1)) 'acos(r1)
      End If
      If (r2<0) Then
        txang = 2*PI-txang 'angle is in 3rd or 4th quadrant
      End If

      'rotate element to 0 degrees but dont rewrite
      tmp(1,1) = rotation(1,1)
      tmp(1,2) = rotation(1,2)
      tmp(1,3) = rotation(1,3)
      tmp(2,1) = rotation(2,1)
      tmp(2,2) = rotation(2,2)
      tmp(2,3) = rotation(2,3)
      tmp(3,1) = rotation(3,1)
      tmp(3,2) = rotation(3,2)
      tmp(3,3) = rotation(3,3)

      rotation(1,2) = tmp(2,1)
      rotation(1,3) = tmp(3,1)
      rotation(2,1) = tmp(1,2)
      rotation(2,3) = tmp(3,2)
      rotation(3,1) = tmp(1,3)
      rotation(3,2) = tmp(2,3)

      status = elem.getOrigin( org )

      If elem.rotate( rotation,org ) = MBE_Success Then

      ' get range box
      If elem.getRange(range) = MBE_Success Then
        x1 = range.xLow  / conv + globalOrg.X
        x2 = range.xHigh / conv + globalOrg.X
        y1 = range.yLow  / conv + globalOrg.Y
        y2 = range.yHigh / conv + globalOrg.Y
      End If

      'if text node work through each line and get max. coords
      If elem.Type = MBE_TextNode Then 'must a text node
        x1 = org.x
        x2 = org.x
        y1 = org.y
        y2 = org.y
        Do While elem.nextComponent = MBE_Success
          If elem.getRange(range) = MBE_Success Then
            t1 = range.xLow  / conv + globalOrg.X
            t2 = range.xHigh / conv + globalOrg.X
            t3 = range.yLow  / conv + globalOrg.Y
            t4 = range.yHigh / conv + globalOrg.Y
            If (x1 > t1 ) Then x1 = t1
            If (x2 < t2 ) Then x2 = t2
            If (y1 > t3 ) Then y1 = t3
            If (y2 < t4 ) Then y2 = t4
          End If
        Loop
      End If

      'range box is padded in y dir., so pad in x dir.
      x1 = x1 - txH/2.0
      x2 = x2 + txH/2.0

      'rotate x1,x2,y1,y2 by txtang about org
      'first shift org to 0
      x1 = x1 - org.x
      x2 = x2 - org.x
      y1 = y1 - org.y
      y2 = y2 - org.y

      'from rotation matrix above:
      '  r1 = cos(txang)
      '  r2 = sin(txang)
      'rotate box cords. back to text angle
      p1.x = x1*r1 - y1*r2 'x1*cos(txang) - y1*sin(txang)
      p1.y = x1*r2 + y1*r1 'x1*sin(txang) + y1*cos(txang)
      p1.z = 0
      p2.x = x1*r1 - y2*r2 'x1*cos(txang) - y2*sin(txang)
      p2.y = x1*r2 + y2*r1 'x1*sin(txang) + y2*cos(txang)
      p2.z = 0
      p3.x = x2*r1 - y2*r2 'x2*cos(txang) - y2*sin(txang)
      p3.y = x2*r2 + y2*r1 'x2*sin(txang) + y2*cos(txang)
      p3.z = 0
      p4.x = x2*r1 - y1*r2 'x2*cos(txang) - y1*sin(txang)
      p4.y = x2*r2 + y1*r1 'x2*sin(txang) + y1*cos(txang)
      p4.z = 0

      'shift back to org.
      p1.x = p1.x + org.x
      p1.y = p1.y + org.y
      p2.x = p2.x + org.x
      p2.y = p2.y + org.y
      p3.x = p3.x + org.x
      p3.y = p3.y + org.y
      p4.x = p4.x + org.x
      p4.y = p4.y + org.y

      'file position of selected text
      filePos = elem.fromLocate ()
      'file position of box to be placed
      endOfFilePos = MbeDgnInfo.endOfFile
      'place box

      MbeSendCommand "place Smartline"
      MbeSendDataPoint p1
      MbeSendDataPoint p2
      MbeSendDataPoint p3
      MbeSendDataPoint p4
      MbeSendDataPoint p1

      'Create a graphic group of text & box
      MbeSendCommand "LOCK GGROUP ON"
      MbeSendCommand "GROUP ADD"
      MbeLocateElement filePos
      MbeLocateElement endOfFilePos
      MbeLocateElement endOfFilePos
      MbeSendReset

    End If
  End If
End If
End Sub


'---------------------------------------------------------------
'
'   Main - main subroutine
'
'---------------------------------------------------------------
Sub Main
    ' set MicroStation to a neutral state
    MbeSendCommand "NULL"

    locate_selectElementsToModify

    MbeWriteStatus "Exiting TextBox"

    MbeWritePrompt
    MbeWriteCommand
End Sub
